
import copy
import json
from typing import List
from dataclasses import asdict
from fastapi import WebSocket
from peacasso.datamodel import GeneratorConfig, SocketData
from peacasso.generator import ImageGenerator
from peacasso.utils import base64_to_pil, pil_to_base64, sanitize_config
import traceback


# enable web socket connection for real time updates
class ConnectionManager:
    """Connection manager for websockets"""

    def __init__(self):
        self.active_connections: List[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)

    async def send_personal_message(self, message: str, websocket: WebSocket):
        await websocket.send_text(message)

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await connection.send_text(message)


async def process_request(request: SocketData, generator: ImageGenerator, websocket: WebSocket):
    """Process socket request"""

    if request.type == "generate":
        prompt_config = GeneratorConfig(**request.data["config"])
        sanitized_config = asdict(sanitize_config(copy.deepcopy(prompt_config)))

        # def generator_callback(i: int, t: int, latents: torch.FloatTensor, images: PIL.Image.Image):
        #     """Callback for generator"""

        #     asyncio.get_event_loop().create_task(websocket.send_text(json.dumps(
        #         {"type": "generator_progress", "data": {"i": i, "config": sanitized_config}})))

        # generator.generate(request.data, socket=socket)
        if prompt_config.init_image is not None:
            prompt_config.init_image, _ = base64_to_pil(prompt_config.init_image)
        if prompt_config.mask_image:
            _, prompt_config.mask_image = base64_to_pil(prompt_config.mask_image)
        response = None
        # prompt_config.callback = generator_callback
        try:
            # print("prompt_config >>>>>> ", prompt_config)
            result = generator.generate(prompt_config)
            images = []
            for image in result["images"]:
                # convert pil image to base64 and prepend with data uri
                images.append("data:image/png;base64, " + pil_to_base64(image))
            result["images"] = images
            # print(result)
            response = json.dumps({"type": "generate_complete", "data": {
                "status": {"status": True, "message": "success"},
                "config": sanitized_config, "result": result}})  # send result to client
        except Exception as e:
            # print("error: {}".format(e))
            traceback.print_exc()
            response = json.dumps({"type": "generate_complete", "data": {
                "status": {"status": False, "message": str(e)},
                "config": sanitized_config, "result": None}})
        await websocket.send_text(response)

    elif request.type == "ping":
        response = json.dumps({"type": "ping", "data": "pong"})
        await websocket.send_text(response)
        # generator.cancel()



import base64
from typing import List
import os
from PIL import Image
import io
import PIL
import torch
import numpy as np

from peacasso.datamodel import GeneratorConfig


def get_dirs(path: str) -> List[str]:
    return next(os.walk(path))[1]


def base64_to_pil(base64_string: str) -> Image:
    # base64_string = base64_string.replace("data:image/png;base64,", "")
    base64_string = base64_string.split(",")[1]
    img_bytes = base64.b64decode(base64_string)
    img = Image.open(io.BytesIO(img_bytes))
    # print number of channels
    mask = None
    if img.mode == "RGBA":
        mask = img.getchannel("A")
        img = img.convert("RGB")
    return img, mask


def pil_to_base64(img: Image) -> str:
    img_byte_arr = io.BytesIO()
    img.save(img_byte_arr, format="PNG")
    img.close()
    return base64.b64encode(img_byte_arr.getvalue()).decode("utf-8")


def sanitize_config(config: GeneratorConfig) -> GeneratorConfig:
    """Sanitize config, remove non serializable fields"""

    if config.init_image:
        config.init_image = None
    if config.mask_image:
        config.mask_image = None
    config.callback = None
    return config


def preprocess_image(image):
    w, h = image.size
    w, h = map(lambda x: x - x % 64, (w, h))  # resize to integer multiple of 32
    image = image.resize((w, h), resample=PIL.Image.LANCZOS)
    image = np.array(image).astype(np.float32) / 255.0
    image = image[None].transpose(0, 3, 1, 2)
    image = torch.from_numpy(image)
    return 2.0 * image - 1.0


def preprocess_mask(mask):
    w, h = mask.size
    w, h = map(lambda x: x - x % 32, (w, h))  # resize to integer multiple of 32
    mask = mask.resize((w // 8, h // 8), resample=PIL.Image.NEAREST)
    mask = np.array(mask).astype(np.float32) / 255.0
    mask = np.tile(mask, (4, 1, 1))
    mask = mask[None].transpose(0, 1, 2, 3)  # what does this step do?
    # mask = 1 - mask  # repaint white, keep black
    mask = torch.from_numpy(mask)
    return mask


def slerp(t, v0, v1, DOT_THRESHOLD=0.9995):
    """ helper function to spherically interpolate two arrays v1 v2 """

    with torch.no_grad():
        if not isinstance(v0, np.ndarray):
            inputs_are_torch = True
            input_device = v0.device
            v0 = v0.cpu().numpy()
            v1 = v1.cpu().numpy()

        dot = np.sum(v0 * v1 / (np.linalg.norm(v0) * np.linalg.norm(v1)))
        if np.abs(dot) > DOT_THRESHOLD:
            v2 = (1 - t) * v0 + t * v1
        else:
            theta_0 = np.arccos(dot)
            sin_theta_0 = np.sin(theta_0)
            theta_t = theta_0 * t
            sin_theta_t = np.sin(theta_t)
            s0 = np.sin(theta_0 - theta_t) / sin_theta_0
            s1 = sin_theta_t / sin_theta_0
            v2 = s0 * v0 + s1 * v1

        if inputs_are_torch:
            v2 = torch.from_numpy(v2).to(input_device)

        return v2



# from dataclasses import dataclass
from dataclasses import field
from random import seed
from typing import Any, List, Optional, Union
from pydantic.dataclasses import dataclass
from typer import Option
import os


@dataclass
class ModelConfig:
    """ Configuration for the HF Diffuser Model"""
    model: Optional[str] = "runwayml/stable-diffusion-v1-5"
    token: Optional[str] = os.environ.get("HF_API_TOKEN")
    device: Optional[str] = "cuda:0"
    revision: Optional[str] = "fp16"


@dataclass
class GeneratorConfig:
    """Configuration for a generation"""

    prompt: Union[str, List[str]]
    num_images: int = 1
    height: Optional[int] = 512
    width: Optional[int] = 512
    num_inference_steps: Optional[int] = 50
    guidance_scale: Optional[float] = 7.5
    eta: Optional[float] = 0.0
    # generator: Optional[Any] = None
    output_type: Optional[str] = "pil"
    strength: float = 0.8
    init_image: Optional[Any] = None
    seed: Optional[Union[int, None]] = None  # e.g. 2147483647
    return_intermediates: bool = False
    mask_image: Optional[Any] = None
    attention_slice: Optional[Union[str, int]] = None
    negative_prompt: Optional[Union[str, List[str]]] = None
    latents: Optional[Any] = None
    callback: Optional[Any] = None
    prompt_weights: Optional[List[float]] = None
    use_prompt_weights: bool = False
    application: Optional[Any] = None
    text_embeddings: Optional[Any] = None
    filter_nsfw: bool = True


@dataclass
class SocketData:
    """Data sent over the websocket"""
    data: Any
    type: str
    token: Optional[str] = None


@dataclass
class ModelResponse:
    """Response from the model"""
    status: bool
    message: Optional[str] = None
    data: Optional[Any] = None



import json
import logging
from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from fastapi.staticfiles import StaticFiles
import os
from peacasso.generator import ImageGenerator
from fastapi.middleware.cors import CORSMiddleware
from peacasso.datamodel import ModelConfig, SocketData
from peacasso.web.backend.processor import ConnectionManager, process_request
import os

logger = logging.getLogger("peacasso")

assert os.environ.get("HF_API_TOKEN") is not None, "HF_API_TOKEN not set"

# load model using env variables
model_config = ModelConfig(
    model=os.environ.get("PEACASSO_MODEL", "runwayml/stable-diffusion-v1-5"),
    revision=os.environ.get("PEACASSO_REVISION", "fp16"),
    device=os.environ.get("PEACASSO_DEVICE", "cuda:0"),
    token=os.environ.get("HF_API_TOKEN")
)
logger.info(
    ">> Loading Imagenerator pipeline with config. " + str(model_config))
generator = ImageGenerator(model_config)
logger.info(">> Imagenerator pipeline loaded.")

app = FastAPI()
# allow cross origin requests for testing on localhost: 800 * ports only
app.add_middleware(
    CORSMiddleware,
    allow_origins=["http://localhost:8000", "http://localhost:8001"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
api = FastAPI(root_path="/api")
app.mount("/api", api)


root_file_path = os.path.dirname(os.path.abspath(__file__))
static_folder_root = os.path.join(root_file_path, "ui")
files_static_root = os.path.join(root_file_path, "files/")

os.makedirs(files_static_root, exist_ok=True)

# mount peacasso front end UI files
app.mount("/", StaticFiles(directory=static_folder_root, html=True), name="ui")
api.mount("/files", StaticFiles(directory=files_static_root, html=True), name="files")


manager = ConnectionManager()


@api.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await manager.connect(websocket)
    try:
        while True:
            data = await websocket.receive_text()
            try:
                request = SocketData(**json.loads(data))
                await process_request(request, generator, websocket)
            except Exception as e:
                print("error: {}".format(e))
                response = json.dumps({"type": "generate_complete", "data": {
                    "status": {"status": False, "message": str("{}".format(e))},
                }})
                await websocket.send_text(response)
                continue

    except WebSocketDisconnect:
        manager.disconnect(websocket)
        await manager.broadcast(f"Client  left the chat")



from dataclasses import asdict
import logging
from PIL import Image
from typing import List, Optional
# from diffusers import StableDiffusionPipeline

import os
import torch
import time

from peacasso.datamodel import GeneratorConfig, ModelConfig
from peacasso.applications import Runner
from peacasso.pipelines import StableDiffusionPipeline

logger = logging.getLogger(__name__)


class ImageGenerator:
    """Class to handle generating image from prompt"""

    def __init__(
        self,
        model_config: ModelConfig
    ) -> None:

        self.create_model(model_config)

    def create_model(self, model_config: ModelConfig):
        # assert model_config. token is not None, "HF_API_TOKEN environment variable must be set."

        device = model_config.device if torch.cuda.is_available() else "cpu"
        self.pipe = StableDiffusionPipeline.from_pretrained(
            model_config.model,
            revision=model_config.revision,
            torch_dtype=torch.float16,
            use_auth_token=model_config.token
        ).to(device)

        self.runner = Runner()

    def generate(self, config: GeneratorConfig) -> Image:
        """Generate image from prompt"""
        return self.runner.run(config, self.pipe)

    def reload(self, model_config) -> None:
        """Clear pipe models from memory, reload pipe with new parameters"""
        del self.pipe
        torch.cuda.empty_cache()
        self.create_model(model_config)

    def list_cuda(self) -> List[int]:
        """List available cuda devices
        Returns:
            List[int]: List of available cuda devices
        """
        available_gpus = [i for i in range(torch.cuda.device_count())]
        return available_gpus
